# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""A brax system."""

import functools
from typing import Tuple

from brax.physics import actuators
from brax.physics import bodies
from brax.physics import colliders
from brax.physics import config_pb2
from brax.physics import integrators
from brax.physics import joints
from brax.physics import math
from brax.physics import pytree
from brax.physics import tree
from brax.physics.base import Info, P, QP, euler_to_quat, validate_config, vec_to_np
import jax
import jax.numpy as jnp


@pytree.register
class System:
  """A brax system."""

  __pytree_ignore__ = ('config', 'num_bodies', 'num_joints', 'num_joint_dof',
                       'num_actuators')

  def __init__(self, config: config_pb2.Config):
    self.config = validate_config(config)

    self.num_bodies = len(config.bodies)
    self.body = bodies.Body(config)

    self.active_pos = 1. * jnp.logical_not(
        jnp.array([vec_to_np(b.frozen.position) for b in config.bodies]))
    self.active_rot = 1. * jnp.logical_not(
        jnp.array([vec_to_np(b.frozen.rotation) for b in config.bodies]))

    self.colliders = colliders.get(self.config, self.body)

    self.num_joints = len(config.joints)
    self.joints = joints.get(self.config, self.body)

    self.num_actuators = len(config.actuators)
    self.num_joint_dof = sum(len(j.angle_limit) for j in config.joints)

    self.actuators = actuators.get(self.config, self.joints)

  @functools.partial(jax.jit, static_argnums=(0, 1))
  def default_qp(self, default_index: int = 0) -> QP:
    """Returns a default state for the system."""
    qps = {}

    # do we have default overrides in the config?
    default_angles, default_qps = {}, {}
    if default_index < len(self.config.defaults):
      defaults = self.config.defaults[default_index]
      default_qps = {qp.name: qp for qp in defaults.qps}
      default_angles = {angle.name: angle for angle in defaults.angles}

    # make the kinematic tree of bodies
    root = tree.Node.from_config(self.config)

    parent_joint = {j.child: j for j in self.config.joints}
    rot_axes = jax.jit(jax.vmap(math.rotate, in_axes=[0, None]))

    for body in root.depth_first():
      if body.name in default_qps:
        # set qp if found in defaults.qps override
        qp = default_qps[body.name]
        qps[body.name] = QP(
            pos=vec_to_np(qp.pos),
            rot=euler_to_quat(qp.rot),
            vel=vec_to_np(qp.vel),
            ang=vec_to_np(qp.ang))
      elif body.name in parent_joint:
        # pos/rot can be determined if the body is the child of a joint
        joint = parent_joint[body.name]
        axes = rot_axes(jnp.eye(3), euler_to_quat(joint.rotation))
        if joint.name in default_angles:
          angles = vec_to_np(default_angles[joint.name].angle) * jnp.pi / 180
        else:
          angles = [(l.min + l.max) * jnp.pi / 360 for l in joint.angle_limit]

        # for each joint angle, rotate by that angle.
        # these are euler intrinsic rotations, so the axes are rotated too
        local_rot = jnp.array([1., 0., 0., 0.])
        for axis, angle in zip(axes, angles):
          axis_rotated = math.rotate(axis, local_rot)
          next_rot = math.quat_rot_axis(axis_rotated, angle)
          local_rot = math.qmult(next_rot, local_rot)
        base_rot = euler_to_quat(joint.reference_rotation)
        local_rot = math.qmult(base_rot, local_rot)

        local_offset = math.rotate(vec_to_np(joint.child_offset), local_rot)
        local_pos = vec_to_np(joint.parent_offset) - local_offset

        parent_qp = qps[joint.parent]
        rot = math.qmult(parent_qp.rot, local_rot)
        pos = parent_qp.pos + math.rotate(local_pos, parent_qp.rot)
        qps[body.name] = QP(
            pos=pos, rot=rot, vel=jnp.zeros(3), ang=jnp.zeros(3))
      else:
        qps[body.name] = QP.zero()

    # any trees that have no body qp overrides in the config are set just above
    # the xy plane.  this convenience operation may be removed in the future.
    body_map = {body.name: body for body in self.config.bodies}
    for node in root.children:
      children = [node.name] + [n.name for n in node.depth_first()]
      if any(c in default_qps for c in children):
        continue  # ignore a tree if some part of it is overriden

      zs = jnp.array([bodies.min_z(qps[c], body_map[c]) for c in children])
      min_z = jnp.min(zs)
      for body in children:
        qp = qps[body]
        pos = qp.pos - min_z * jnp.array([0., 0., 1.])
        qps[body] = qp.replace(pos=pos)

    qps = [qps[body.name] for body in self.config.bodies]
    return jax.tree_multimap((lambda *args: jnp.stack(args)), *qps)

  @functools.partial(jax.jit, static_argnums=(0,))
  def info(self, qp: QP) -> Info:
    """Return info about a system state."""
    zero = P(jnp.zeros((self.num_bodies, 3)), jnp.zeros((self.num_bodies, 3)))

    dp_c = sum([c.apply(qp) for c in self.colliders], zero)
    dp_j = sum([j.apply(qp) for j in self.joints], zero)
    info = Info(dp_c, dp_j, zero)
    return info

  @functools.partial(jax.jit, static_argnums=(0,))
  def step(self, qp: QP, act: jnp.ndarray) -> Tuple[QP, Info]:
    """Calculates a physics step for a system, returns next state and info."""

    def substep(carry, _):
      qp, info = carry
      dt = self.config.dt / self.config.substeps

      # apply kinetic step
      qp = integrators.kinetic(self.config, qp, dt, self.active_pos,
                               self.active_rot)

      # apply impulses arising from joints and actuators
      zero = P(jnp.zeros((self.num_bodies, 3)), jnp.zeros((self.num_bodies, 3)))
      dp_j = sum([j.apply(qp) for j in self.joints], zero)
      dp_a = sum([a.apply(qp, act) for a in self.actuators], zero)
      qp = integrators.potential(self.config, qp, dp_j + dp_a, dt,
                                 self.active_pos, self.active_rot)

      # apply collision velocity updates
      dp_c = sum([c.apply(qp) for c in self.colliders], zero)
      qp = integrators.potential_collision(self.config, qp, dp_c,
                                           self.active_pos, self.active_rot)

      info = Info(info.contact + dp_c, info.joint + dp_j, info.actuator + dp_a)

      return (qp, info), ()

    # update collider statistics for culling
    for c in self.colliders:
      c.cull.update(qp)

    zero = P(jnp.zeros((self.num_bodies, 3)), jnp.zeros((self.num_bodies, 3)))
    info = Info(contact=zero, joint=zero, actuator=zero)
    (qp, info), _ = jax.lax.scan(substep, (qp, info), (), self.config.substeps)
    return qp, info



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trains a 2D walker to run in the +x direction."""

from typing import Tuple

from brax.envs import hopper


class Walker2d(hopper.Hopper):
  """Trains a 2D walker to run in the +x direction.

  This is similar to the Walker2d-V3 Mujoco environment in OpenAI Gym, which is
  a variant of Hopper with two legs. The two legs do not collide with each
  other.
  """

  def __init__(self,
               *argv,
               healthy_z_range: Tuple[float, float] = (0.7, 2.0),
               **kwargs):
    super().__init__(
        *argv,
        **kwargs,
        healthy_z_range=healthy_z_range,
        system_config=_SYSTEM_CONFIG)


_SYSTEM_CONFIG = """
bodies {
  name: "torso"
  colliders {
    position {}
    rotation {}
    capsule {
      radius: 0.05
      length: 0.5
    }
  }
  inertia { x: 1.0 y: 1.0 z: 1.0 }
  mass: 3.6651914
}
bodies {
  name: "thigh"
  colliders {
    position { z: -0.225 }
    rotation {}
    capsule {
      radius: 0.05
      length: 0.55
    }
  }
  inertia { x: 1.0 y: 1.0 z: 1.0 }
  mass: 4.0578904
}
bodies {
  name: "leg"
  colliders {
    position {}
    rotation {}
    capsule {
      radius: 0.04
      length: 0.58
    }
  }
  inertia { x: 1.0 y: 1.0 z: 1.0 }
  mass: 2.7813568
}
bodies {
  name: "foot"
  colliders {
    position {
      x: -0.1
      y: -0.2
      z: -0.1
    }
    rotation { y: 90.0 }
    capsule {
      radius: 0.06
      length: 0.32
    }
  }
  inertia { x: 1.0 y: 1.0 z: 1.0 }
  mass: 3.1667254
}
bodies {
  name: "thigh_left"
  colliders {
    position { z: -0.225 }
    rotation {}
    capsule {
      radius: 0.05
      length: 0.55
    }
  }
  inertia { x: 1.0 y: 1.0 z: 1.0 }
  mass: 4.0578904
}
bodies {
  name: "leg_left"
  colliders {
    position {}
    rotation {}
    capsule {
      radius: 0.04
      length: 0.58
    }
  }
  inertia { x: 1.0 y: 1.0 z: 1.0 }
  mass: 2.7813568
}
bodies {
  name: "foot_left"
  colliders {
    position {
      x: -0.1
      y: -0.2
      z: -0.1
    }
    rotation { y: 90.0 }
    capsule {
      radius: 0.06
      length: 0.32
    }
  }
  inertia { x: 1.0 y: 1.0 z: 1.0 }
  mass: 3.1667254
}
bodies {
  name: "floor"
  colliders {
    plane {}
  }
  inertia { x: 1.0 y: 1.0 z: 1.0 }
  frozen { all: true }
}
joints {
  name: "thigh_joint"
  stiffness: 10000.0
  parent: "torso"
  child: "thigh"
  parent_offset { z: -0.2 }
  rotation { z: -90.0 }
  angle_limit { min: -150.0 }
  angular_damping: 20.0
}
joints {
  name: "leg_joint"
  stiffness: 10000.0
  parent: "thigh"
  child: "leg"
  parent_offset { z: -0.45 }
  child_offset { z: 0.25 }
  rotation { z: -90.0 }
  angle_limit { min: -150.0 }
  angular_damping: 20.0
}
joints {
  name: "foot_joint"
  stiffness: 10000.0
  parent: "leg"
  child: "foot"
  parent_offset { z: -0.25 }
  child_offset {
    x: -0.2
    y: -0.2
    z: -0.1
  }
  rotation { z: -90.0 }
  angle_limit { min: -45.0 max: 45.0 }
  angular_damping: 20.0
}
joints {
  name: "thigh_left_joint"
  stiffness: 10000.0
  parent: "torso"
  child: "thigh_left"
  parent_offset { z: -0.2 }
  rotation { z: -90.0 }
  angle_limit { min: -150.0 }
  angular_damping: 20.0
}
joints {
  name: "leg_left_joint"
  stiffness: 10000.0
  parent: "thigh_left"
  child: "leg_left"
  parent_offset { z: -0.45 }
  child_offset { z: 0.25 }
  rotation { z: -90.0 }
  angle_limit { min: -150.0 }
  angular_damping: 20.0
}
joints {
  name: "foot_left_joint"
  stiffness: 10000.0
  parent: "leg_left"
  child: "foot_left"
  parent_offset { z: -0.25 }
  child_offset {
    x: -0.2
    y: -0.2
    z: -0.1
  }
  rotation { z: -90.0 }
  angle_limit { min: -45.0 max: 45.0 }
  angular_damping: 20.0
}
actuators {
  name: "thigh_joint"
  joint: "thigh_joint"
  strength: 100.0
  torque {}
}
actuators {
  name: "leg_joint"
  joint: "leg_joint"
  strength: 100.0
  torque {}
}
actuators {
  name: "foot_joint"
  joint: "foot_joint"
  strength: 100.0
  torque {}
}
actuators {
  name: "thigh_left_joint"
  joint: "thigh_left_joint"
  strength: 100.0
  torque {}
}
actuators {
  name: "leg_left_joint"
  joint: "leg_left_joint"
  strength: 100.0
  torque {}
}
actuators {
  name: "foot_left_joint"
  joint: "foot_left_joint"
  strength: 100.0
  torque {}
}
friction: 0.9
gravity { z: -9.81 }
velocity_damping: 1.0
angular_damping: -0.05
baumgarte_erp: 0.1
collide_include {
  first: "floor"
  second: "torso"
}
collide_include {
  first: "floor"
  second: "thigh"
}
collide_include {
  first: "floor"
  second: "leg"
}
collide_include {
  first: "floor"
  second: "foot"
}
collide_include {
  first: "floor"
  second: "thigh_left"
}
collide_include {
  first: "floor"
  second: "leg_left"
}
collide_include {
  first: "floor"
  second: "foot_left"
}
dt: 0.02
substeps: 4
frozen {
  position { y: 1.0 }
  rotation { x: 1.0 z: 1.0 }
}
defaults {
  qps { name: "torso" pos { z: 1.19 } }
  angles { name: "thigh_joint" angle {} }
  angles { name: "leg_joint" angle {} }
  angles { name: "thigh_left_joint" angle {} }
  angles { name: "leg_left_joint" angle {} }
}
"""



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Colliders push apart bodies that are in contact."""

import abc
import itertools
from typing import Any, Callable, List, Tuple
import warnings

from brax.physics import bodies
from brax.physics import config_pb2
from brax.physics import math
from brax.physics import pytree
from brax.physics.base import P, QP, euler_to_quat, take, vec_to_np
import jax
import jax.numpy as jnp


class Collidable:
  """Part of a body (with geometry and mass/inertia) that can collide.

  Collidables can repeat for a geometry. e.g. a body with a box collider has 8
  corner collidables.
  """

  def __init__(self, collidables: List[config_pb2.Body], body: bodies.Body):
    self.body = take(body, jnp.array([body.index[c.name] for c in collidables]))
    self.pos = jnp.array(
        [vec_to_np(c.colliders[0].position) for c in collidables])

  def position(self, qp: QP) -> jnp.ndarray:
    """Returns the collidable's position in world space."""
    pos = jnp.take(qp.pos, self.body.idx, axis=0, mode='clip')
    rot = jnp.take(qp.rot, self.body.idx, axis=0, mode='clip')
    return pos + jax.vmap(math.rotate)(self.pos, rot)


@pytree.register
class Contact:
  """Stores information about contacts between two collidables."""

  def __init__(self, pos: jnp.array, vel: jnp.array, normal: jnp.array,
               penetration: jnp.array):
    """Creates a Contact.

    Args:
      pos: contact position in world space
      vel: contact velocity in world space
      normal: normal vector of surface providing contact
      penetration: distance the two collidables are penetrating one-another
    """
    self.pos = pos
    self.vel = vel
    self.normal = normal
    self.penetration = penetration


class Cull(abc.ABC):
  """Selects collidable pair candidates for collision detection."""

  @abc.abstractmethod
  def get(self) -> Tuple[Collidable, Collidable]:
    """Returns collidable pair candidates for collision detection."""

  def update(self, qp: QP):
    """Updates candidate statistics given qp world state."""


@pytree.register
class AllPairs(Cull):
  """Naive strategy: returns all possible pairs for collision detection."""

  def __init__(self, col_a: Collidable, col_b: Collidable,
               mask: List[Tuple[int, int]]):
    self.col_a = take(col_a, jnp.array([a for a, _ in mask]))
    self.col_b = take(col_b, jnp.array([b for _, b in mask]))

  def get(self) -> Tuple[Collidable, Collidable]:
    return self.col_a, self.col_b


@pytree.register
class NearNeighbors(Cull):
  """Returns top K nearest neighbor collidables for collision detection."""

  __pytree_ignore__ = ('cutoff',)

  def __init__(self, col_a: Collidable, col_b: Collidable,
               mask: List[Tuple[int, int]], cutoff: int):
    dist_off = jnp.zeros(col_a.body.idx.shape + col_b.body.idx.shape)
    # TODO: consider accounting for bounds/radius of a collidable
    dist_mask = dist_off + float('inf')
    mask = (jnp.array([a for a, _ in mask]), jnp.array([b for _, b in mask]))
    dist_off += dist_mask.at[mask].set(0)
    self.dist_off = dist_off
    self.cutoff = cutoff
    self.candidate_a, self.candidate_b = col_a, col_b
    self.col_a, self.col_b = self.candidate_a, self.candidate_b

  def update(self, qp: QP):
    if self.candidate_a is self.candidate_b:
      pos = self.candidate_a.position(qp)
      dist = jax.vmap(lambda pt: jnp.linalg.norm(pos - pt, axis=-1))(pos)
    else:
      pos_a = self.candidate_a.position(qp)
      pos_b = self.candidate_b.position(qp)
      dist = jax.vmap(lambda pt: jnp.linalg.norm(pos_b - pt, axis=-1))(pos_a)
    # add in offset and convert to closeness
    sim = -(dist + self.dist_off)
    # TODO: return a status if more valid candidates than cutoff
    _, idx = jax.lax.top_k(sim.ravel(), self.cutoff)
    self.col_a = take(self.candidate_a, idx // sim.shape[-1])
    self.col_b = take(self.candidate_b, idx % sim.shape[-1])

  def get(self) -> Tuple[Collidable, Collidable]:
    return self.col_a, self.col_b


class Collider(abc.ABC):
  """Calculates impulses given contacts from a contact function."""

  __pytree_ignore__ = ('contact_fn', 'cull', 'friction', 'elasticity',
                       'baumgarte_erp')

  def __init__(self, contact_fn: Callable[[Any, Any, QP, QP], Contact],
               cull: Cull, config: config_pb2.Config):
    """Creates a PairwiseCollider that exhaustively checks for contacts.

    Args:
      contact_fn: a function that returns contacts given collidables and qp
      cull: a cull strategy
      config: for specifying global system config
    """
    self.contact_fn = contact_fn
    self.cull = cull
    self.friction = config.friction
    self.elasticity = config.elasticity
    self.baumgarte_erp = config.baumgarte_erp * config.substeps / config.dt

  @abc.abstractmethod
  def apply(self, qp: QP) -> P:
    """Returns impulse from any potential contacts between collidables.

    Args:
      qp: Coordinate/velocity frame of the bodies.

    Returns:
      dP: Impulse to apply to the bodies in the collision.
    """


@pytree.register
class OneWayCollider(Collider):
  """Calculates one-way impulses, where the second collidable is static."""

  def apply(self, qp: QP) -> P:
    """Returns impulse from any potential contacts between collidables."""
    col_a, col_b = self.cull.get()
    qp_a = take(qp, col_a.body.idx)
    qp_b = take(qp, col_b.body.idx)
    contact = jax.vmap(self.contact_fn)(col_a, col_b, qp_a, qp_b)
    dp = jax.vmap(self._contact)(qp_a, col_a, contact)

    contact = jnp.where(jnp.any(dp.vel, axis=-1), 1.0, 0.0)
    contact = jax.ops.segment_sum(contact, col_a.body.idx, qp.pos.shape[0])
    dp_vel = jax.ops.segment_sum(dp.vel, col_a.body.idx, qp.pos.shape[0])
    dp_ang = jax.ops.segment_sum(dp.ang, col_a.body.idx, qp.pos.shape[0])

    # equally distribute impulse over possible contacts
    dp_vel = dp_vel / jnp.reshape(1e-8 + contact, (dp_vel.shape[0], 1))
    dp_ang = dp_ang / jnp.reshape(1e-8 + contact, (dp_ang.shape[0], 1))

    return P(vel=dp_vel, ang=dp_ang)

  def _contact(self, qp: QP, col: Collidable, contact: Contact) -> P:
    """Calculates impulse on a body due to a contact."""
    rel_pos = contact.pos - qp.pos
    baumgarte_vel = self.baumgarte_erp * contact.penetration
    normal_vel = jnp.dot(contact.normal, contact.vel)
    temp1 = jnp.matmul(col.body.inertia, jnp.cross(rel_pos, contact.normal))
    ang = jnp.dot(contact.normal, jnp.cross(temp1, rel_pos))
    impulse = (-1. * (1. + self.elasticity) * normal_vel + baumgarte_vel) / (
        (1. / col.body.mass) + ang)
    dp_n = col.body.impulse(qp, impulse * contact.normal, contact.pos)

    # apply drag due to friction acting parallel to the surface contact
    vel_d = contact.vel - normal_vel * contact.normal
    impulse_d = math.safe_norm(vel_d) / ((1. / (col.body.mass)) + ang)
    # drag magnitude cannot exceed max friction
    impulse_d = jnp.minimum(impulse_d, self.friction * impulse)
    dir_d = vel_d / (1e-6 + math.safe_norm(vel_d))
    dp_d = col.body.impulse(qp, -impulse_d * dir_d, contact.pos)
    # apply collision if penetrating, approaching, and oriented correctly
    apply_n = jnp.where(
        (contact.penetration > 0.) & (normal_vel < 0) & (impulse > 0.), 1., 0.)
    # apply drag if moving laterally above threshold
    apply_d = apply_n * jnp.where(math.safe_norm(vel_d) > 0.01, 1., 0.)

    return dp_n * apply_n + dp_d * apply_d


@pytree.register
class TwoWayCollider(Collider):
  """Calculates two-way impulses on collidable pairs."""

  def apply(self, qp: QP) -> P:
    """Returns impulse from any potential contacts between collidables."""
    col_a, col_b = self.cull.get()
    qp_a = take(qp, col_a.body.idx)
    qp_b = take(qp, col_b.body.idx)
    contact = jax.vmap(self.contact_fn)(col_a, col_b, qp_a, qp_b)
    dp_a, dp_b = jax.vmap(self._contact)(col_a, col_b, qp_a, qp_b, contact)

    body_idx = jnp.concatenate((col_a.body.idx, col_b.body.idx))
    dp_vel = jnp.concatenate((dp_a.vel, dp_b.vel))
    dp_ang = jnp.concatenate((dp_a.ang, dp_b.ang))
    contact = jnp.where(jnp.any(dp_vel, axis=-1), 1.0, 0.0)
    contact = jax.ops.segment_sum(contact, body_idx, qp.pos.shape[0])
    dp_vel = jax.ops.segment_sum(dp_vel, body_idx, qp.pos.shape[0])
    dp_ang = jax.ops.segment_sum(dp_ang, body_idx, qp.pos.shape[0])

    # equally distribute impulse over possible contacts
    contact = jnp.reshape(1e-8 + contact, (dp_vel.shape[0], 1))
    dp_vel = dp_vel / contact
    dp_ang = dp_ang / contact
    return P(vel=dp_vel, ang=dp_ang)

  def _contact(self, col_a: Collidable, col_b: Collidable, qp_a: QP, qp_b: QP,
               contact: Contact) -> Tuple[P, P]:
    """Calculates impulse on a body due to a contact."""
    rel_pos_a = contact.pos - qp_a.pos
    rel_pos_b = contact.pos - qp_b.pos
    baumgarte_vel = self.baumgarte_erp * contact.penetration
    normal_vel = jnp.dot(contact.normal, contact.vel)
    temp1 = jnp.matmul(col_a.body.inertia, jnp.cross(rel_pos_a, contact.normal))
    temp2 = jnp.matmul(col_b.body.inertia, jnp.cross(rel_pos_b, contact.normal))
    ang = jnp.dot(contact.normal,
                  jnp.cross(temp1, rel_pos_a) + jnp.cross(temp2, rel_pos_b))
    impulse = (-1. * (1. + self.elasticity) * normal_vel + baumgarte_vel) / (
        (1. / col_a.body.mass) + (1. / col_b.body.mass) + ang)
    dp_n_a = col_a.body.impulse(qp_a, -impulse * contact.normal, contact.pos)
    dp_n_b = col_b.body.impulse(qp_b, impulse * contact.normal, contact.pos)

    # apply drag due to friction acting parallel to the surface contact
    vel_d = contact.vel - normal_vel * contact.normal
    impulse_d = math.safe_norm(vel_d) / ((1. / col_a.body.mass) +
                                         (1. / col_b.body.mass) + ang)
    # drag magnitude cannot exceed max friction
    impulse_d = jnp.minimum(impulse_d, self.friction * impulse)
    dir_d = vel_d / (1e-6 + math.safe_norm(vel_d))
    dp_d_a = col_a.body.impulse(qp_a, impulse_d * dir_d, contact.pos)
    dp_d_b = col_a.body.impulse(qp_b, -impulse_d * dir_d, contact.pos)
    # apply collision normal if penetrating, approaching, and oriented correctly
    apply_n = jnp.where(
        (contact.penetration > 0.) & (normal_vel < 0) & (impulse > 0.), 1., 0.)
    # apply drag if moving laterally above threshold
    apply_d = apply_n * jnp.where(math.safe_norm(vel_d) > 0.01, 1., 0.)

    dp_a = dp_n_a * apply_n + dp_d_a * apply_d
    dp_b = dp_n_b * apply_n + dp_d_b * apply_d
    return dp_a, dp_b


@pytree.register
class BoxCorner(Collidable):
  """A box corner."""

  def __init__(self, boxes: List[config_pb2.Body], body: bodies.Body):
    super().__init__([boxes[i // 8] for i in range(len(boxes) * 8)], body)
    coords = jnp.array(list(itertools.product((-1, 1), (-1, 1), (-1, 1))))
    corners = []
    for b in boxes:
      col = b.colliders[0]
      rot = euler_to_quat(col.rotation)
      box = coords * vec_to_np(col.box.halfsize)
      box = jax.vmap(math.rotate, in_axes=(0, None))(box, rot)
      box = box + vec_to_np(col.position)
      corners.extend(box)
    self.corner = jnp.array(corners)


@pytree.register
class Plane(Collidable):
  """An infinite plane with normal pointing in the +z direction."""


@pytree.register
class Capsule(Collidable):
  """A capsule with an ends pointing in the +z, -z directions."""

  def __init__(self, capsules: List[config_pb2.Body], body: bodies.Body):
    super().__init__(capsules, body)
    ends = []
    radii = []
    for c in capsules:
      col = c.colliders[0]
      axis = math.rotate(jnp.array([0., 0., 1.]), euler_to_quat(col.rotation))
      segment_length = col.capsule.length / 2. - col.capsule.radius
      ends.append(axis * segment_length)
      radii.append(col.capsule.radius)
    self.end = jnp.array(ends)
    self.radius = jnp.array(radii)


@pytree.register
class CapsuleEnd(Collidable):
  """A capsule with variable ends either in the +z or -z directions."""

  def __init__(self, capsules: List[config_pb2.Body], body: bodies.Body):
    var_caps = [[c] if c.colliders[0].capsule.end else [c, c] for c in capsules]
    super().__init__(sum(var_caps, []), body)
    ends = []
    radii = []
    for c in capsules:
      col = c.colliders[0]
      axis = math.rotate(jnp.array([0., 0., 1.]), euler_to_quat(col.rotation))
      segment_length = col.capsule.length / 2. - col.capsule.radius
      for end in [col.capsule.end] if col.capsule.end else [-1, 1]:
        ends.append(vec_to_np(col.position) + end * axis * segment_length)
        radii.append(col.capsule.radius)
    self.end = jnp.array(ends)
    self.radius = jnp.array(radii)


@pytree.register
class HeightMap(Collidable):
  """A height map with heights in a grid layout."""

  def __init__(self, heightmaps: List[config_pb2.Body], body: bodies.Body):
    super().__init__(heightmaps, body)
    heights = []
    cell_sizes = []
    for h in heightmaps:
      col = h.colliders[0]
      mesh_size = int(jnp.sqrt(len(col.heightMap.data)))
      if len(col.heightMap.data) != mesh_size**2:
        raise ValueError('height map data length should be a perfect square.')
      height = jnp.array(col.heightMap.data).reshape((mesh_size, mesh_size))
      heights.append(height)
      cell_sizes.append(col.heightMap.size / (mesh_size - 1))
    self.height = jnp.array(heights)
    self.cell_size = jnp.array(cell_sizes)


def box_plane(box: BoxCorner, _: Plane, qp_a: QP, qp_b: QP) -> Contact:
  """Returns contact between a box corner and a plane."""
  pos, vel = math.to_world(qp_a, box.corner)
  normal = math.rotate(jnp.array([0., 0., 1.]), qp_b.rot)
  penetration = jnp.dot(qp_b.pos - pos, normal)
  return Contact(pos, vel, normal, penetration)


def box_heightmap(box: BoxCorner, hm: HeightMap, qp_a: QP, qp_b: QP) -> Contact:
  """Returns contact between a box corner and a height map."""
  # Note that this only checks box corners against height map surfaces, and is
  # missing box planes against height map points.
  pos, vel = math.to_world(qp_a, box.corner)
  pos = math.inv_rotate(pos - qp_b.pos, qp_b.rot)
  uv_pos = pos[:2] / hm.cell_size
  # find the square in the mesh that enclose the candidate point, with mesh
  # indices ux_idx, ux_udx_u, uv_idx_v, uv_idx_uv.
  uv_idx = jnp.floor(uv_pos).astype(jnp.int32)
  uv_idx_u = uv_idx + jnp.array([1, 0], dtype=jnp.int32)
  uv_idx_v = uv_idx + jnp.array([0, 1], dtype=jnp.int32)
  uv_idx_uv = uv_idx + jnp.array([1, 1], dtype=jnp.int32)
  # find the orientation of the triangle of this square that encloses the
  # candidate point
  delta_uv = uv_pos - uv_idx
  # whether the corner lies on the first or secound triangle:
  mu = jnp.where(delta_uv[0] + delta_uv[1] < 1, 1, -1)
  # compute the mesh indices of the vertices of this triangle
  p0 = jnp.where(delta_uv[0] + delta_uv[1] < 1, uv_idx, uv_idx_uv)
  p1 = jnp.where(delta_uv[0] + delta_uv[1] < 1, uv_idx_u, uv_idx_v)
  p2 = jnp.where(delta_uv[0] + delta_uv[1] < 1, uv_idx_v, uv_idx_u)
  h0 = hm.height[p0[0], p0[1]]
  h1 = hm.height[p1[0], p1[1]]
  h2 = hm.height[p2[0], p2[1]]

  raw_normal = jnp.array(
      [-mu * (h1 - h0), -mu * (h2 - h0), hm.cell_size])
  normal = raw_normal / jnp.linalg.norm(raw_normal)
  normal = math.rotate(normal, qp_b.rot)
  height = jnp.array(
      [p0[0] * hm.cell_size, p0[1] * hm.cell_size, h0])
  penetration = jnp.dot(height - pos, normal)
  return Contact(pos, vel, normal, penetration)


def capsule_plane(cap: CapsuleEnd, _: Plane, qp_a: QP, qp_b: QP) -> Contact:
  """Returns contact between a capsule and a plane."""
  cap_end_world = qp_a.pos + math.rotate(cap.end, qp_a.rot)
  normal = math.rotate(jnp.array([0., 0., 1.]), qp_b.rot)
  pos = cap_end_world - normal * cap.radius
  vel = qp_a.vel + jnp.cross(qp_a.ang, pos - qp_a.pos)
  penetration = jnp.dot(qp_b.pos - pos, normal)
  return Contact(pos, vel, normal, penetration)


def capsule_capsule(cap_a: Capsule, cap_b: Capsule, qp_a: QP,
                    qp_b: QP) -> Contact:
  """Returns contact between two capsules."""
  def endpoints(end, qp, offset):
    pos = qp.pos + math.rotate(offset, qp.rot)
    end = math.rotate(end, qp.rot)
    return pos + end, pos - end

  def closest_segment_point(a, b, pt):
    ab = b - a
    t = jnp.dot(pt - a, ab) / (jnp.dot(ab, ab) + 1e-10)
    return a + jnp.clip(t, 0., 1.) * ab

  a0, a1 = endpoints(cap_a.end, qp_a, cap_a.pos)
  b0, b1 = endpoints(cap_b.end, qp_b, cap_b.pos)
  v0, v1, v2, v3 = b0 - a0, b1 - a0, b0 - a1, b1 - a1
  d0, d1 = jnp.dot(v0, v0), jnp.dot(v1, v1)
  d2, d3 = jnp.dot(v2, v2), jnp.dot(v3, v3)
  a_best = jnp.where((d2 < d0) | (d2 < d1) | (d3 < d0) | (d3 < d1), a1, a0)
  b_best = closest_segment_point(b0, b1, a_best)
  a_best = closest_segment_point(a0, a1, b_best)

  penetration_vec = b_best - a_best
  dist = math.safe_norm(penetration_vec)
  normal = penetration_vec / (1e-6 + dist)
  penetration = cap_a.radius + cap_b.radius - dist
  pos = (a_best + b_best) / 2
  vel = math.world_velocity(qp_b, pos) - math.world_velocity(qp_a, pos)
  return Contact(pos, vel, normal, penetration)


def get(config: config_pb2.Config, body: bodies.Body) -> List[Collider]:
  """Creates all colliders given a config."""
  def key_fn(x, y):
    return tuple(sorted((body.index.get(x, -1), body.index.get(y, -1))))
  include = {key_fn(f.first, f.second) for f in config.collide_include}
  # exclude colliders for joint parents and children, unless explicitly included
  ignore = {key_fn(j.parent, j.child) for j in config.joints} - include
  # exclude colliders where both bodies are frozen
  frozen = [b.name for b in config.bodies if b.frozen.all]
  ignore.union({key_fn(x, y) for x, y in itertools.combinations(frozen, 2)})

  # flatten and emit one collider per body
  flat_bodies = []
  for b in config.bodies:
    for collider in b.colliders:
      # we treat spheres as sphere-shaped capsules with a single end
      if collider.WhichOneof('type') == 'sphere':
        radius = collider.sphere.radius
        collider = config_pb2.Collider()
        collider.capsule.radius = radius
        collider.capsule.length = 2 * radius
        collider.capsule.end = 1

      new_body = config_pb2.Body()
      new_body.CopyFrom(b)
      new_body.ClearField('colliders')
      new_body.colliders.append(collider)
      flat_bodies.append(new_body)

  # group by type
  type_colliders = {}
  for b in flat_bodies:
    key = b.colliders[0].WhichOneof('type')
    if key not in type_colliders:
      type_colliders[key] = []
    type_colliders[key].append(b)

  # add colliders
  supported_types = {
      ('box', 'plane'): (BoxCorner, Plane, box_plane),
      ('box', 'heightMap'): (BoxCorner, HeightMap, box_heightmap),
      ('capsule', 'plane'): (CapsuleEnd, Plane, capsule_plane),
      ('capsule', 'capsule'): (Capsule, Capsule, capsule_capsule),
  }
  supported_near_neighbors = {('capsule', 'capsule')}
  collidable_cache = {}

  ret = []
  for type_a, type_b in itertools.combinations_with_replacement(
      type_colliders, 2):
    type_a, type_b = sorted((type_a, type_b))

    # calculate all possible body pairs that can collide via these two types
    cols_a, cols_b = type_colliders[type_a], type_colliders[type_b]
    if type_a == type_b:
      candidates = itertools.combinations(cols_a, 2)
    else:
      candidates = itertools.product(cols_a, cols_b)
    pairs = [(body.index[a.name], body.index[b.name]) for a, b in candidates]
    if include:
      pairs = [x for x in pairs if tuple(sorted(x)) in include]
    pairs = [x for x in pairs if tuple(sorted(x)) not in ignore]

    if not pairs:
      continue

    if (type_a, type_b) not in supported_types:
      warnings.warn(f'unsupported collider pair: {type_a}, {type_b}')
      continue

    # create our collidables
    col_cls_a, col_cls_b, contact_fn = supported_types[(type_a, type_b)]
    if col_cls_a not in collidable_cache:
      collidable_cache[col_cls_a] = col_cls_a(cols_a, body)
    if col_cls_b not in collidable_cache:
      collidable_cache[col_cls_b] = col_cls_b(cols_b, body)
    col_a = collidable_cache[col_cls_a]
    col_b = collidable_cache[col_cls_b]

    # convert pairs from body idx to collidable idx
    body_to_collidable_a, body_to_collidable_b = {}, {}
    for i, body_idx in enumerate(col_a.body.idx):
      body_to_collidable_a.setdefault(body_idx, []).append(i)
    for i, body_idx in enumerate(col_b.body.idx):
      body_to_collidable_b.setdefault(body_idx, []).append(i)
    mask = []
    for body_idx_a, body_idx_b in pairs:
      ias = body_to_collidable_a[body_idx_a]
      ibs = body_to_collidable_b[body_idx_b]
      for ia, ib in itertools.product(ias, ibs):
        mask.append((ia, ib))

    if config.collider_cutoff and len(pairs) > config.collider_cutoff and (
        type_a, type_b) in supported_near_neighbors:
      cull = NearNeighbors(col_a, col_b, mask, config.collider_cutoff)
    else:
      cull = AllPairs(col_a, col_b, mask)
    if all(b.frozen.all for b in cols_b):
      collider = OneWayCollider(contact_fn, cull, config)
    else:
      collider = TwoWayCollider(contact_fn, cull, config)

    ret.append(collider)

  return ret



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Joints connect bodies and constrain their movement."""

import abc
from typing import Any, List, Tuple

from brax.physics import bodies
from brax.physics import config_pb2
from brax.physics import math
from brax.physics import pytree
from brax.physics.base import P, QP, euler_to_quat, take, vec_to_np
import jax
import jax.numpy as jnp


class Joint(abc.ABC):
  """A joint connects two bodies and constrains their movement.

  This constraint is determined by axes that define how to bodies may move in
  relation to one-another.
  """

  __pytree_ignore__ = ('index', 'dof')

  def __init__(self,
               joints: List[config_pb2.Joint],
               body: bodies.Body,
               spring_damping_coeff: float = 2.0):
    """Creates a Joint that connects two bodies and constrains their movement.

    Args:
      joints: list of joints (all of the same type) to batch together
      body: batched body that contain the parents and children of each joint
      spring_damping_coeff: coefficient for setting default spring damping
    """
    self.stiffness = jnp.array([j.stiffness for j in joints])
    self.angular_damping = jnp.array([j.angular_damping for j in joints])
    self.spring_damping = jnp.array([
        j.spring_damping if j.HasField('spring_damping') else
        spring_damping_coeff * jnp.sqrt(j.stiffness) for j in joints
    ])
    self.limit_strength = jnp.array([
        j.limit_strength if j.HasField('limit_strength') else j.stiffness
        for j in joints
    ])
    self.limit = jnp.array([[[i.min, i.max]
                             for i in j.angle_limit]
                            for j in joints]) / 180.0 * jnp.pi
    self.body_p = take(body, jnp.array([body.index[j.parent] for j in joints]))
    self.body_c = take(body, jnp.array([body.index[j.child] for j in joints]))
    self.off_p = jnp.array([vec_to_np(j.parent_offset) for j in joints])
    self.off_c = jnp.array([vec_to_np(j.child_offset) for j in joints])
    self.index = {j.name: i for i, j in enumerate(joints)}
    self.dof = len(joints[0].angle_limit)
    v_rot = jax.vmap(math.rotate, in_axes=[0, None])
    relative_quats = jnp.array(
        [euler_to_quat(j.reference_rotation) for j in joints])
    self.axis_c = jnp.array(
        [v_rot(jnp.eye(3), euler_to_quat(j.rotation)) for j in joints])
    self.axis_p = jnp.array(
        [v_rot(j, r) for j, r in zip(self.axis_c, relative_quats)])

  def apply(self, qp: QP) -> P:
    """Returns impulses to constrain and align bodies connected by a joint.

    Args:
      qp: State data for system

    Returns:
      dP: Impulses on all bodies to maintain joint constraints
    """
    qp_p = take(qp, self.body_p.idx)
    qp_c = take(qp, self.body_c.idx)
    dp_p, dp_c = self.apply_reduced(qp_p, qp_c)

    # sum together all impulse contributions across parents and children
    body_idx = jnp.concatenate((self.body_p.idx, self.body_c.idx))
    dp_vel = jnp.concatenate((dp_p.vel, dp_c.vel))
    dp_ang = jnp.concatenate((dp_p.ang, dp_c.ang))
    dp_vel = jax.ops.segment_sum(dp_vel, body_idx, qp.pos.shape[0])
    dp_ang = jax.ops.segment_sum(dp_ang, body_idx, qp.pos.shape[0])

    return P(vel=dp_vel, ang=dp_ang)

  def angle_vel(self, qp: QP) -> Tuple[Any, Any]:
    """Returns joint angle and velocity.

    Args:
      qp: State data for system

    Returns:
      angle: n-tuple of joint angles where n = # DoF of the joint
      vel: n-tuple of joint velocities where n = # DoF of the joint
    """

    @jax.vmap
    def op(joint, qp_p, qp_c):
      axes, angles = joint.axis_angle(qp_p, qp_c)
      vels = tuple([jnp.dot(qp_p.ang - qp_c.ang, axis) for axis in axes])
      return angles, vels

    qp_p = take(qp, self.body_p.idx)
    qp_c = take(qp, self.body_c.idx)
    angles, vels = op(self, qp_p, qp_c)

    return angles, vels

  @abc.abstractmethod
  def apply_reduced(self, qp_p: QP, qp_c: QP) -> Tuple[P, P]:
    """Returns impulses to constrain and align bodies connected by a joint.

    Operates in reduced joint space.

    Args:
      qp_p: Joint parent state data
      qp_c: Joint child state data

    Returns:
      dp_p: Joint parent impulse
      dp_c: Joint child impulse
    """

  @abc.abstractmethod
  def axis_angle(self, qp_p: QP, qp_c: QP) -> Tuple[Any, Any]:
    """Returns axes and angles of a single joint.

    vmap across axis_angle to get all joints.

    Args:
      qp_p: State for parent body
      qp_c: State for child body

    Returns:
      axis: n-tuple of joint axes where n = # DoF of the joint
      angle: n-tuple of joint angles where n = # DoF of the joint
    """


@pytree.register
class Revolute(Joint):
  """A revolute joint constrains two bodies around a single axis.

  Constructs a revolute joint where the parent's local x-axis is constrained
  to point in the same direction as the child's local x-axis.  This construction
  follows the line of nodes convention shared by the universal and spherical
  joints for x-y'-z'' intrinsic euler angles.
  """

  def __init__(self, joints: List[config_pb2.Joint], body: bodies.Body):
    super().__init__(joints, body, .5)

  @jax.vmap
  def apply_reduced(self, qp_p: QP, qp_c: QP) -> Tuple[P, P]:
    """Returns calculated impulses in compressed joint space."""
    pos_p, vel_p = math.to_world(qp_p, self.off_p)
    pos_c, vel_c = math.to_world(qp_c, self.off_c)

    # push the bodies towards their offsets
    # TODO: cap this damping so that it can't overcorrect
    impulse = (pos_p - pos_c) * self.stiffness + self.spring_damping * (
        vel_p - vel_c)
    dp_p = self.body_p.impulse(qp_p, -impulse, pos_p)
    dp_c = self.body_c.impulse(qp_c, impulse, pos_c)

    # torque the bodies to align their axes
    (axis,), (angle,) = self.axis_angle(qp_p, qp_c)
    axis_c = math.rotate(self.axis_c[0], qp_c.rot)
    torque = self.stiffness * jnp.cross(axis, axis_c)

    # torque the bodies to stay within angle limits
    dang = jnp.where(angle < self.limit[0][0], self.limit[0][0] - angle, 0)
    dang = jnp.where(angle > self.limit[0][1], self.limit[0][1] - angle, dang)
    torque -= self.limit_strength * axis * dang

    # damp the angular motion
    torque -= self.angular_damping * (qp_p.ang - qp_c.ang)

    dang_p = jnp.matmul(self.body_p.inertia, torque)
    dang_c = jnp.matmul(self.body_c.inertia, -torque)
    dp_p = dp_p.replace(ang=dp_p.ang + dang_p)
    dp_c = dp_c.replace(ang=dp_c.ang + dang_c)

    return dp_p, dp_c

  def axis_angle(self, qp_p: QP, qp_c: QP) -> Tuple[Any, Any]:
    """Returns axes and angles of a single joint."""
    axis_p = math.rotate(self.axis_p[0], qp_p.rot)
    ref_p = math.rotate(self.axis_p[2], qp_p.rot)
    ref_c = math.rotate(self.axis_c[2], qp_c.rot)
    # algebraically the same as the calculation in `Spherical`, but simpler
    # because child local-x and parent local-x are constrained to be the same
    psi = math.signed_angle(axis_p, ref_p, ref_c)
    return (axis_p,), (psi,)


@pytree.register
class Universal(Joint):
  """A revolute joint constrains two bodies around two axes.

  Constructs a universal joint defined as the first two degrees of freedom
  of a spherical joint.  See `Spherical` for details.
  """

  @jax.vmap
  def apply_reduced(self, qp_p: QP, qp_c: QP) -> Tuple[P, P]:
    """Returns calculated impulses in compressed joint space."""
    pos_p, vel_p = math.to_world(qp_p, self.off_p)
    pos_c, vel_c = math.to_world(qp_c, self.off_c)

    # push the bodies towards their offsets
    # TODO: cap this damping so that it can't overcorrect
    impulse = (pos_p - pos_c) * self.stiffness + self.spring_damping * (
        vel_p - vel_c)
    dp_p = self.body_p.impulse(qp_p, -impulse, pos_p)
    dp_c = self.body_c.impulse(qp_c, impulse, pos_c)

    # torque the bodies to align to a joint plane
    (axis_1, axis_2), (angle_1, angle_2) = self.axis_angle(qp_p, qp_c)
    axis_c_proj = axis_2 - jnp.dot(axis_2, axis_1) * axis_1
    axis_c_proj = axis_c_proj / math.safe_norm(axis_c_proj)
    torque = (self.limit_strength / 5.) * jnp.cross(axis_c_proj, axis_2)

    # torque the bodies to stay within angle limits
    limit_1, limit_2 = self.limit
    dang_1 = jnp.where(angle_1 < limit_1[0], limit_1[0] - angle_1, 0)
    dang_1 = jnp.where(angle_1 > limit_1[1], limit_1[1] - angle_1, dang_1)
    dang_2 = jnp.where(angle_2 < limit_2[0], limit_2[0] - angle_2, 0)
    dang_2 = jnp.where(angle_2 > limit_2[1], limit_2[1] - angle_2, dang_2)
    torque -= self.limit_strength * (axis_1 * dang_1 + axis_2 * dang_2)

    # damp the angular motion
    torque -= self.angular_damping * (qp_p.ang - qp_c.ang)

    dang_p = jnp.matmul(self.body_p.inertia, torque)
    dang_c = jnp.matmul(self.body_c.inertia, -torque)
    dp_p = dp_p.replace(ang=dp_p.ang + dang_p)
    dp_c = dp_c.replace(ang=dp_c.ang + dang_c)

    return dp_p, dp_c

  def axis_angle(self, qp_p: QP, qp_c: QP) -> Tuple[Any, Any]:
    """Returns axes and angles of a single joint."""
    v_rot = jax.vmap(math.rotate, in_axes=[0, None])
    axis_p_rotated = v_rot(self.axis_p, qp_p.rot)
    axis_c_rotated = v_rot(self.axis_c, qp_c.rot)
    axis_1_p = axis_p_rotated[0]
    axis_2_p = axis_p_rotated[1]
    axis_1_c = axis_c_rotated[0]
    axis_2_c = axis_c_rotated[1]
    axis_3_c = axis_c_rotated[2]

    line_of_nodes = jnp.cross(axis_3_c, axis_1_p)
    line_of_nodes = line_of_nodes / (1e-10 + math.safe_norm(line_of_nodes))

    y_N_normal = axis_1_p

    psi = math.signed_angle(y_N_normal, axis_2_p, line_of_nodes)

    axis_1_p_in_xz_c = jnp.dot(axis_1_p, axis_1_c) * axis_1_c + jnp.dot(
        axis_1_p, axis_2_c) * axis_2_c
    axis_1_p_in_xz_c = axis_1_p_in_xz_c / (1e-10 +
                                           math.safe_norm(axis_1_p_in_xz_c))
    theta = jnp.arccos(jnp.clip(jnp.dot(axis_1_p_in_xz_c, axis_1_p), -1,
                                1)) * jnp.sign(jnp.dot(axis_1_p, axis_3_c))

    axis = (axis_1_p, axis_2_c)
    angle = (psi, theta)

    return axis, angle


@pytree.register
class Spherical(Joint):
  """A spherical joint constrains two bodies around three axes.

  Constructs a spherical joint which returns intrinsic euler angles in the
    x-y'-z'' convention between the parent and child.  Uses the line of nodes
    construction described in section 3.2.3.2 here:
    https://www.sedris.org/wg8home/Documents/WG80485.pdf
    """

  @jax.vmap
  def apply_reduced(self, qp_p: QP, qp_c: QP) -> Tuple[P, P]:
    """Returns calculated impulses in compressed joint space."""
    pos_p, vel_p = math.to_world(qp_p, self.off_p)
    pos_c, vel_c = math.to_world(qp_c, self.off_c)

    # push the bodies towards their offsets
    # TODO: cap this damping so that it can't overcorrect
    impulse = (pos_p - pos_c) * self.stiffness + self.spring_damping * (
        vel_p - vel_c)
    dp_p = self.body_p.impulse(qp_p, -impulse, pos_p)
    dp_c = self.body_c.impulse(qp_c, impulse, pos_c)

    # torque the bodies to stay within angle limits
    axis, angle = self.axis_angle(qp_p, qp_c)
    angle_1, angle_2, angle_3 = angle
    limit_1, limit_2, limit_3 = self.limit
    dang_1 = jnp.where(angle_1 < limit_1[0], limit_1[0] - angle_1, 0)
    dang_1 = jnp.where(angle_1 > limit_1[1], limit_1[1] - angle_1, dang_1)
    dang_2 = jnp.where(angle_2 < limit_2[0], limit_2[0] - angle_2, 0)
    dang_2 = jnp.where(angle_2 > limit_2[1], limit_2[1] - angle_2, dang_2)
    dang_3 = jnp.where(angle_3 < limit_3[0], limit_3[0] - angle_3, 0)
    dang_3 = jnp.where(angle_3 > limit_3[1], limit_3[1] - angle_3, dang_3)

    # TODO: fully decouple different torque axes
    torque = axis[0] * dang_1 + axis[1] * dang_2 + axis[2] * dang_3
    torque *= self.limit_strength * -1.

    # damp the angular motion
    torque -= self.angular_damping * (qp_p.ang - qp_c.ang)

    dp_p = dp_p.replace(ang=dp_p.ang + jnp.matmul(self.body_p.inertia, torque))
    dp_c = dp_c.replace(ang=dp_c.ang + jnp.matmul(self.body_c.inertia, -torque))

    return dp_p, dp_c

  def axis_angle(self, qp_p: QP, qp_c: QP) -> Tuple[Any, Any]:
    """Returns axes and angles of a single joint."""
    v_rot = jax.vmap(math.rotate, in_axes=[0, None])
    axis_p_rotated = v_rot(self.axis_p, qp_p.rot)
    axis_c_rotated = v_rot(self.axis_c, qp_c.rot)
    axis_1_p = axis_p_rotated[0]
    axis_2_p = axis_p_rotated[1]
    axis_1_c = axis_c_rotated[0]
    axis_2_c = axis_c_rotated[1]
    axis_3_c = axis_c_rotated[2]

    line_of_nodes = jnp.cross(axis_3_c, axis_1_p)
    line_of_nodes = line_of_nodes / (1e-10 + math.safe_norm(line_of_nodes))

    y_N_normal = axis_1_p

    psi = math.signed_angle(y_N_normal, axis_2_p, line_of_nodes)

    axis_1_p_in_xz_c = jnp.dot(axis_1_p, axis_1_c) * axis_1_c + jnp.dot(
        axis_1_p, axis_2_c) * axis_2_c
    axis_1_p_in_xz_c = axis_1_p_in_xz_c / (1e-10 +
                                           math.safe_norm(axis_1_p_in_xz_c))

    ang_between_1_p_xz_c = jnp.dot(axis_1_p_in_xz_c, axis_1_p)
    theta = jnp.arccos(jnp.clip(ang_between_1_p_xz_c, -1, 1)) * jnp.sign(
        jnp.dot(axis_1_p, axis_3_c))
    yc_N_normal = -1. * axis_3_c

    phi = math.signed_angle(yc_N_normal, axis_2_c, line_of_nodes)

    axis = (axis_1_p, axis_2_c, axis_3_c)
    angle = (psi, theta, phi)

    return axis, angle


def get(config: config_pb2.Config, body: bodies.Body) -> List[Joint]:
  """Creates all joints given a config."""
  joints = {}
  for joint in config.joints:
    dof = len(joint.angle_limit)
    if dof not in joints:
      joints[dof] = []
    joints[dof].append(joint)

  # ensure stable order for joint application: dof
  joints = sorted(joints.items(), key=lambda kv: kv[0])
  ret = []
  for k, v in joints:
    if k == 1:
      ret.append(Revolute(v, body))
    elif k == 2:
      ret.append(Universal(v, body))
    elif k == 3:
      ret.append(Spherical(v, body))
    else:
      raise RuntimeError(f'invalid number of joint limits: {k}')

  return ret
